"""Toolforge etcdctl module tests."""
from typing import Optional
from unittest import TestCase, mock

from ClusterShell.MsgTree import MsgTreeElem
from cumin import Config, nodeset

from spicerack.remote import RemoteHosts
from spicerack.toolforge.etcdctl import EtcdctlController, HealthStatus, TooManyHosts, UnableToParseOutput


def _assert_called_with_single_param(param: str, mock_obj: mock.MagicMock) -> None:
    mock_obj.assert_called()
    call_count = 0
    for call in mock_obj.mock_calls:
        for arg in call[1]:
            if isinstance(arg, str) and param in arg:
                call_count += 1

    assert (
        call_count == 1
    ), f"Expected to have 1 call with param {param}, but got {call_count}, all calls: {mock_obj.mock_calls}"


def _assert_not_called_with_single_param(param: str, mock_obj: mock.MagicMock) -> None:
    mock_obj.assert_called_once()
    for call in mock_obj.mock_calls:
        for arg in call[1]:
            if isinstance(arg, str) and param in arg:
                assert False, f"Parameter {param} was passed on a call to {mock_obj}: {call}"


def _get_mock_run_sync(
    return_value: Optional[bytes] = None, side_effect: Optional[list[bytes]] = None
) -> mock.MagicMock:
    if side_effect is not None:
        return mock.MagicMock(
            side_effect=[
                (
                    iter(
                        [
                            ("test0.local.host", MsgTreeElem(return_value, parent=MsgTreeElem())),
                        ]
                    )
                )
                for return_value in side_effect
            ]
        )

    return mock.MagicMock(return_value=iter([("test0.local.host", MsgTreeElem(return_value, parent=MsgTreeElem()))]))


class TestEtcdctlController(TestCase):
    """TestEtcdctlController."""

    def test_raises_if_more_than_one_node_is_used(self):
        """Test that raises if more than one node is used."""
        nodes = RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test[0,1].local.host"))

        with self.assertRaises(TooManyHosts):
            EtcdctlController(remote_host=nodes)


class TestGetHealth(TestCase):
    """TestGetHealth."""

    def test_passes_correct_cert_file(self):
        """Test that passes correct cert file by default."""
        expected_cert_file = "/etc/etcd/ssl/test0.local.host.pem"
        mock_run_sync = _get_mock_run_sync(
            return_value=b"""
                member 415090d15def9053 is healthy: got healthy result from https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2379
                cluster is healthy
            """,  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            controller.get_cluster_health()

        _assert_called_with_single_param(
            param=f"--cert-file {expected_cert_file}",
            mock_obj=mock_run_sync,
        )

    def test_passes_correct_ca_file(self):
        """Test that passes correct ca file by default."""
        expected_ca_file = "/etc/etcd/ssl/ca.pem"
        mock_run_sync = _get_mock_run_sync(
            return_value=b"""
                member 415090d15def9053 is healthy: got healthy result from https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2379
                cluster is healthy
            """,  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            controller.get_cluster_health()

        _assert_called_with_single_param(
            param=f"--ca-file {expected_ca_file}",
            mock_obj=mock_run_sync,
        )

    def test_passes_correct_key_file(self):
        """Test that passes correct key file by default."""
        expected_key_file = "/etc/etcd/ssl/test0.local.host.priv"
        mock_run_sync = _get_mock_run_sync(
            return_value=b"""
                member 415090d15def9053 is healthy: got healthy result from https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2379
                cluster is healthy
            """,  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            controller.get_cluster_health()

        _assert_called_with_single_param(
            param=f"--key-file {expected_key_file}",
            mock_obj=mock_run_sync,
        )

    def test_passes_correct_endpoints(self):
        """Test that passes correct endpoints by default."""
        expected_endpoints = "https://test0.local.host:2379"
        mock_run_sync = _get_mock_run_sync(
            return_value=b"""
                member 415090d15def9053 is healthy: got healthy result from https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2379
                cluster is healthy
            """,  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            controller.get_cluster_health()

        _assert_called_with_single_param(
            param=f"--endpoints {expected_endpoints}",
            mock_obj=mock_run_sync,
        )

    def test_parses_result_with_one_member(self):
        """Test that parses result with one member."""
        expected_members = {"415090d15def9053": HealthStatus.healthy}
        mock_run_sync = _get_mock_run_sync(
            return_value=b"""
                member 415090d15def9053 is healthy: got healthy result from https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2379
                cluster is healthy
            """,  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            gotten_result = controller.get_cluster_health()

        mock_run_sync.assert_called_once()
        assert gotten_result.members_status == expected_members

    def test_parses_result_with_many_members(self):
        """Test that parses result with many members."""
        expected_members = {
            "415090d15def9053": HealthStatus.healthy,
            "5208bbf5c00e7cdf": HealthStatus.healthy,
        }
        mock_run_sync = _get_mock_run_sync(
            return_value=b"""
                member 415090d15def9053 is healthy: got healthy result from https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2379
                member 5208bbf5c00e7cdf is healthy: got healthy result from https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2379
                cluster is healthy
            """,  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            gotten_result = controller.get_cluster_health()

        mock_run_sync.assert_called_once()
        assert expected_members == gotten_result.members_status

    def test_parses_result_with_member_down(self):
        """Test that parses result with member down."""
        expected_members = {
            "415090d15def9053": HealthStatus.healthy,
            "5208bbf5c00e7cdf": HealthStatus.unhealthy,
        }
        mock_run_sync = _get_mock_run_sync(
            return_value=b"""
                member 415090d15def9053 is healthy: got healthy result from https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2379
                member 5208bbf5c00e7cdf is unhealthy: got unhealthy result from https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2379
                cluster is healthy
            """,  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            gotten_result = controller.get_cluster_health()

        mock_run_sync.assert_called_once()
        assert expected_members == gotten_result.members_status

    def test_gets_cluster_healthy(self):
        """Test that parses cluster global status when healthy."""
        expected_global_status = HealthStatus.healthy
        mock_run_sync = _get_mock_run_sync(
            return_value=b"""
                member 5208bbf5c00e7cdf is unhealthy: got unhealthy result from https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2379
                cluster is healthy
            """,  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            gotten_result = controller.get_cluster_health()

        mock_run_sync.assert_called_once()
        assert expected_global_status == gotten_result.global_status

    def test_gets_cluster_unhealthy(self):
        """Test that parses cluster global status when unhealthy."""
        expected_global_status = HealthStatus.unhealthy
        mock_run_sync = _get_mock_run_sync(
            return_value=b"""
                member 5208bbf5c00e7cdf is unhealthy: got unhealthy result from https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2379
                cluster is unhealthy
            """,  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            gotten_result = controller.get_cluster_health()

        mock_run_sync.assert_called_once()
        assert expected_global_status == gotten_result.global_status

    def test_raises_when_no_global_cluster_health(self):
        """Test that parses cluster global status when unhealthy."""
        mock_run_sync = _get_mock_run_sync(
            return_value=b"""
                member 5208bbf5c00e7cdf is unhealthy: got unhealthy result from https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2379
            """,  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            with self.assertRaises(UnableToParseOutput):
                controller.get_cluster_health()

        mock_run_sync.assert_called_once()


class TestGetClusterInfo(TestCase):
    """TestGetClusterInfo."""

    def test_passes_correct_cert_file(self):
        """Test that passes correct cert file by default."""
        expected_cert_file = "/etc/etcd/ssl/test0.local.host.pem"
        mock_run_sync = _get_mock_run_sync(return_value=b"")
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            controller.get_cluster_info()

        _assert_called_with_single_param(
            param=f"--cert-file {expected_cert_file}",
            mock_obj=mock_run_sync,
        )

    def test_passes_correct_ca_file(self):
        """Test that passes correct ca file by default."""
        expected_ca_file = "/etc/etcd/ssl/ca.pem"
        mock_run_sync = _get_mock_run_sync(return_value=b"")
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            controller.get_cluster_info()

        _assert_called_with_single_param(
            param=f"--ca-file {expected_ca_file}",
            mock_obj=mock_run_sync,
        )

    def test_passes_correct_key_file(self):
        """Test that passes correct key file by default."""
        expected_key_file = "/etc/etcd/ssl/test0.local.host.priv"
        mock_run_sync = _get_mock_run_sync(return_value=b"")
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            controller.get_cluster_info()

        _assert_called_with_single_param(
            param=f"--key-file {expected_key_file}",
            mock_obj=mock_run_sync,
        )

    def test_passes_correct_endpoints(self):
        """Test that passes correct endpoints by default."""
        expected_endpoints = "https://test0.local.host:2379"
        mock_run_sync = _get_mock_run_sync(return_value=b"")
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            controller.get_cluster_info()

        _assert_called_with_single_param(
            param=f"--endpoints {expected_endpoints}",
            mock_obj=mock_run_sync,
        )

    def test_parses_result_with_one_member(self):
        """Test that parses result with one member."""
        expected_node = {
            "member_id": "415090d15def9053",
            "name": "toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud",
            "status": "up",
            "isLeader": True,
            "peerURLs": "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2380",
            "clientURLs": "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2379",
        }
        mock_run_sync = _get_mock_run_sync(
            return_value=b"""
                                415090d15def9053: name=toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud peerURLs=https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2380 clientURLs=https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2379 isLeader=true
                            """,  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            gotten_result = controller.get_cluster_info()

        mock_run_sync.assert_called_once()
        assert len(gotten_result) == 1
        assert expected_node["member_id"] in gotten_result
        assert gotten_result[expected_node["member_id"]] == expected_node

    def test_parses_result_with_many_members(self):
        """Test that parses result with many members."""
        expected_result = {
            "415090d15def9053": {
                "member_id": "415090d15def9053",
                "name": "toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud",
                "status": "up",
                "isLeader": True,
                "peerURLs": "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2380",
                "clientURLs": "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2379",
            },
            "5208bbf5c00e7cdf": {
                "clientURLs": "https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2379",
                "isLeader": False,
                "member_id": "5208bbf5c00e7cdf",
                "name": "toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud",
                "peerURLs": "https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2380",
                "status": "up",
            },
        }
        mock_run_sync = _get_mock_run_sync(
            return_value=b"""
                                415090d15def9053: name=toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud peerURLs=https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2380 clientURLs=https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2379 isLeader=true
                                5208bbf5c00e7cdf: name=toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud peerURLs=https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2380 clientURLs=https://toolsbeta-test-k8s-etcd-6.toolsbeta.eqiad1.wikimedia.cloud:2379 isLeader=false
                            """,  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            gotten_result = controller.get_cluster_info()

        mock_run_sync.assert_called_once()
        assert expected_result == gotten_result

    def test_parses_result_with_member_down(self):
        """Test that parses result with member down."""
        expected_result = {
            "415090d15def9053": {
                "member_id": "415090d15def9053",
                "name": "toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud",
                "status": "up",
                "isLeader": True,
                "peerURLs": "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2380",
                "clientURLs": "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2379",
            },
            "cf612c785df58f6a": {
                "member_id": "cf612c785df58f6a",
                "peerURLs": "https://idontexist.localhost:1234",
                "status": "unstarted",
            },
        }
        mock_run_sync = _get_mock_run_sync(
            return_value=b"""
                415090d15def9053: name=toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud peerURLs=https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2380 clientURLs=https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2379 isLeader=true
                cf612c785df58f6a[unstarted]: peerURLs=https://idontexist.localhost:1234
            """  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            gotten_result = controller.get_cluster_info()

        mock_run_sync.assert_called_once()
        assert expected_result == gotten_result

    def test_raises_when_getting_member_without_id(self):
        """Test that raises when getting member without id."""
        mock_run_sync = _get_mock_run_sync(
            return_value=b"""
                415090d15def9053: name=toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud peerURLs=https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2380 clientURLs=https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2379 isLeader=true
                peerURLs=https://idontexist.localhost:1234
            """  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with self.assertRaises(UnableToParseOutput):
            with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
                controller.get_cluster_info()


class TestEnsureNodeExists(TestCase):
    """TestEnsureNodeExists."""

    def test_skips_addition_if_member_already_exists(self):
        """Test that skips addition if member already exists."""
        existing_member_fqdn = "i.already.exist"
        existing_member_peer_url = f"https://{existing_member_fqdn}:1234"
        expected_member_id = "1234556789012345"
        mock_run_sync = _get_mock_run_sync(
            return_value=f"""
                415090d15def9053: name=toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud peerURLs=https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2380 clientURLs=https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2379 isLeader=true
                {expected_member_id}: name={existing_member_fqdn} peerURLs={existing_member_peer_url}
            """.encode()  # noqa: E501
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            gotten_member_id = controller.ensure_node_exists(
                new_member_fqdn=existing_member_fqdn,
                member_peer_url=existing_member_peer_url,
            )

        _assert_called_with_single_param(param="list", mock_obj=mock_run_sync)
        _assert_not_called_with_single_param(param="add", mock_obj=mock_run_sync)
        assert gotten_member_id == expected_member_id

    def test_updates_the_member_if_the_peer_url_does_not_match(self):
        """Test that updates the member if the peer url does not match."""
        existing_member_fqdn = "i.already.exist"
        existing_member_peer_url = f"https://{existing_member_fqdn}:1234"
        expected_member_id = "1234556789012345"
        mock_run_sync = _get_mock_run_sync(
            side_effect=[
                (
                    "415090d15def9053: name=toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud peerURLs="
                    "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2380 clientURLs="
                    "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2379 isLeader=true\n"
                    f"{expected_member_id}: name={existing_member_fqdn} "
                    f"peerURLs={existing_member_peer_url}_but_different"
                ).encode(),
                b"""Updated :)""",
                (
                    "415090d15def9053: name=toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud peerURLs="
                    "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2380 clientURLs="
                    "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2379 isLeader=false\n"
                    f"{expected_member_id}: name={existing_member_fqdn} peerURLs={existing_member_peer_url}"
                ).encode(),
            ]
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            gotten_member_id = controller.ensure_node_exists(
                new_member_fqdn=existing_member_fqdn,
                member_peer_url=existing_member_peer_url,
            )

        _assert_called_with_single_param(param="update", mock_obj=mock_run_sync)
        assert gotten_member_id == expected_member_id

    def test_adds_the_member_if_not_there(self):
        """Test that adds the member if not there."""
        new_member_fqdn = "i.already.exist"
        new_member_peer_url = f"https://{new_member_fqdn}:1234"
        expected_member_id = "1234556789012345"
        mock_run_sync = _get_mock_run_sync(
            side_effect=[
                (
                    "415090d15def9053: name=toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud peerURLs="
                    "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2380 clientURLs="
                    "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2379 isLeader=true"
                ).encode(),
                b"""Added :)""",
                (
                    "415090d15def9053: name=toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud peerURLs="
                    "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2380 clientURLs="
                    "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2379 isLeader=false\n"
                    f"{expected_member_id}: name={new_member_fqdn} peerURLs={new_member_peer_url}"
                ).encode(),
            ]
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            gotten_member_id = controller.ensure_node_exists(
                new_member_fqdn=new_member_fqdn,
                member_peer_url=new_member_peer_url,
            )

        _assert_called_with_single_param(param="add", mock_obj=mock_run_sync)
        assert gotten_member_id == expected_member_id

    def test_uses_default_member_url_if_not_passed(self):
        """Test that uses default member url if not passed."""
        new_member_fqdn = "i.already.exist"
        expected_peer_url = f"https://{new_member_fqdn}:2380"
        mock_run_sync = _get_mock_run_sync(
            side_effect=[
                b"",
                b"""Added :)""",
                f"""
                    415090d15def9053: name={new_member_fqdn} peerURLs={expected_peer_url}
                """.encode(),
            ]
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            controller.ensure_node_exists(new_member_fqdn=new_member_fqdn)

        _assert_called_with_single_param(param=expected_peer_url, mock_obj=mock_run_sync)


class TestEnsureNodeDoesNotExist(TestCase):
    """TestEnsureNodeDoesNotExist."""

    def test_skips_removal_if_member_does_not_exist(self):
        """Test that skips removal if member does not exist."""
        non_existing_member_fqdn = "i.dont.exist"
        expected_result = None
        mock_run_sync = _get_mock_run_sync(
            return_value=(
                "415090d15def9053: name=toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud peerURLs="
                "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2380 clientURLs="
                "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2379 isLeader=true"
            ).encode()
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            gotten_result = controller.ensure_node_does_not_exist(
                member_fqdn=non_existing_member_fqdn,
            )

        _assert_called_with_single_param(param="list", mock_obj=mock_run_sync)
        _assert_not_called_with_single_param(param="remove", mock_obj=mock_run_sync)
        assert gotten_result == expected_result

    def test_removes_the_member_if_there_already(self):
        """Test that it removes the member if there already."""
        member_fqdn = "i.already.exist"
        expected_member_id = "1234556789012345"
        mock_run_sync = _get_mock_run_sync(
            side_effect=[
                (
                    "415090d15def9053: name=toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud peerURLs="
                    "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2380 clientURLs="
                    "https://toolsbeta-test-k8s-etcd-9.toolsbeta.eqiad1.wikimedia.cloud:2379 isLeader=false\n"
                    f"{expected_member_id}: name={member_fqdn} peerURLs=http://some.url"
                ).encode(),
                "Removed :)",
            ]
        )
        controller = EtcdctlController(
            remote_host=RemoteHosts(config=mock.MagicMock(specset=Config), hosts=nodeset("test0.local.host")),
        )

        with mock.patch.object(RemoteHosts, "run_sync", mock_run_sync):
            gotten_result = controller.ensure_node_does_not_exist(member_fqdn=member_fqdn)

        _assert_called_with_single_param(param="remove", mock_obj=mock_run_sync)
        assert gotten_result == expected_member_id



"""Group3 get argument_parser() raise."""
__title__ = __doc__


def argument_parser():
    """As required by spicerack._cookbook."""
    raise RuntimeError("argument_parser() raise")


def run(_args, _spicerack):
    """As required by spicerack._cookbook."""



# pylint: disable=too-many-lines
"""Puppet module tests."""
import json
import re
from collections import namedtuple
from datetime import datetime, timedelta, timezone
from unittest import mock

import pytest
from ClusterShell.MsgTree import MsgTreeElem
from cumin import nodeset

from spicerack import puppet
from spicerack.administrative import Reason
from spicerack.exceptions import SpicerackError
from spicerack.remote import RemoteExecutionError, RemoteHosts

REASON = Reason("Disable reason", "user1", "orchestration-host", task_id="T12345")
PUPPET_SERVER_CA_CERT_METADATA_SIGNED = (
    b'{"signed": [{"name":"test.example.com","state":"signed","fingerprint":"00:AA",'
    b'"fingerprints":{"default":"00:AA","SHA1":"11:BB","SHA256": "00:AA","SHA512":"22:CC"},'
    b'"dns_alt_names":["DNS:service.example.com"]}]}'
)
PUPPET_SERVER_CA_CERT_METADATA_REQUESTED = (
    b'{"signed": [{"name":"test.example.com","state":"requested","fingerprint":"00:AA",'
    b'"fingerprints":{"default":"00:AA","SHA1":"11:BB","SHA256": "00:AA","SHA512":"22:CC"},'
    b'"dns_alt_names":["DNS:service.example.com"]}]}'
)
PUPPET_SERVER_CA_CERT_METADATA_MISSING = b'{"missing": ["test.example.org"]}'
PUPPET_CA_CERT_METADATA_SIGNED = (
    b'[{"name":"test.example.com","state":"signed","fingerprint":"00:AA",'
    b'"fingerprints":{"default":"00:AA","SHA1":"11:BB","SHA256": "00:AA","SHA512":"22:CC"},'
    b'"dns_alt_names":["DNS:service.example.com"]}]'
)
PUPPET_CA_CERT_METADATA_REQUESTED = (
    b'[{"name":"test.example.com","state":"requested","fingerprint":"00:AA",'
    b'"fingerprints":{"default":"00:AA","SHA1":"11:BB","SHA256": "00:AA","SHA512":"22:CC"},'
    b'"dns_alt_names":["DNS:service.example.com"]}]'
)
PUPPET_GENERATE_CERTIFICATE_SUCCESS = b"""Info: Creating a new SSL key for test.example.com
Info: Caching certificate for ca
Info: csr_attributes file loading from /etc/puppet/csr_attributes.yaml
Info: Creating a new SSL certificate request for test.example.com
Info: Certificate Request fingerprint (SHA256): 00:FF
Info: Caching certificate for ca
Exiting; no certificate found and waitforcert is disabled
"""
PUPPET_GENERATE_CERTIFICATE_FAILED = rb"""Info: Creating a new SSL key for test.example.com
Info: Caching certificate for ca
Info: Caching certificate for test.example.com
Error: Could not request certificate: The certificate retrieved from the master does not match the agent's private key.
Certificate fingerprint: 00:FF
To fix this, remove the certificate from both the master and the agent and then start a puppet run, which will...
On the master:
  puppet cert clean test.example.com
On the agent:
  1a. On most platforms: find /var/lib/puppet/ssl -name test.example.com.pem -delete
  1b. On Windows: del "\var\lib\puppet\ssl\certs\test.example.com.pem" /f
  2. puppet agent -t
Exiting; failed to retrieve certificate and waitforcert is disabled
"""

MockedDnsSrv = namedtuple("MockedDnsSrv", ["target"])
MockedDnsAnswer = namedtuple("MockedDnsAnswer", ["ttl", "rrset"])


@mock.patch("spicerack.puppet.Dns", autospec=True)
def test_get_ca_via_srv_record(mocked_dns_resolve):
    """An instance of Spicerack should allow to get a PuppetServer instance."""
    dns_answer = MockedDnsAnswer(ttl=600, rrset=[MockedDnsSrv(target="puppetserver1001.eqiad.wmnet")])
    mocked_dns_resolve.return_value.resolve.return_value = dns_answer
    assert puppet.get_ca_via_srv_record("eqiad.wmnet") == "puppetserver1001.eqiad.wmnet"


@pytest.mark.parametrize(
    "response, err_msg",
    (
        (
            MockedDnsAnswer(ttl=600, rrset=None),
            "Unable to find record for _x-puppet-ca._tcp.eqiad.wmnet",
        ),
        (
            MockedDnsAnswer(ttl=600, rrset=[]),
            "Unable to find record for _x-puppet-ca._tcp.eqiad.wmnet",
        ),
        (
            MockedDnsAnswer(ttl=600, rrset=[MockedDnsSrv(target="foo"), MockedDnsSrv(target="bar")]),
            "_x-puppet-ca._tcp.eqiad.wmnet returned multiple ca servers: foo,bar",
        ),
    ),
)
@mock.patch("spicerack.puppet.Dns", autospec=True)
def test_get_ca_via_srv_record_raises(mocked_dns_resolve, response, err_msg):
    """An instance of Spicerack should allow to get a PuppetServer instance."""
    mocked_dns_resolve.return_value.resolve.return_value = response
    with pytest.raises(SpicerackError, match=err_msg):
        puppet.get_ca_via_srv_record("eqiad.wmnet")
    mocked_dns_resolve.assert_called_once()


def test_get_puppet_ca_hostname_ok():
    """It should get the hostname of the Puppet CA from the local Puppet agent."""
    assert puppet.get_puppet_ca_hostname() == "puppetmaster1001.eqiad.wmnet"


class TestPuppetHosts:
    """Test class for the PuppetHosts class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_remote_hosts = mock.MagicMock(spec_set=RemoteHosts)
        self.mocked_remote_hosts.__len__.return_value = 1
        self.puppet_hosts = puppet.PuppetHosts(self.mocked_remote_hosts)

    @pytest.mark.parametrize(
        "verbatim_reason, result",
        ((False, f"disable-puppet {REASON.quoted()}"), (True, f'disable-puppet "{REASON.reason}"')),
    )
    def test_disable(self, verbatim_reason, result):
        """It should disable Puppet on the hosts."""
        self.puppet_hosts.disable(REASON, verbatim_reason)
        self.mocked_remote_hosts.run_sync.assert_called_once_with(result)

    @pytest.mark.parametrize(
        "verbatim_reason, result",
        ((False, f"enable-puppet {REASON.quoted()}"), (True, f'enable-puppet "{REASON.reason}"')),
    )
    def test_enable(self, verbatim_reason, result):
        """It should enable Puppet on the hosts."""
        self.puppet_hosts.enable(REASON, verbatim_reason)
        self.mocked_remote_hosts.run_sync.assert_called_once_with(result)

    @pytest.mark.parametrize("verbatim_reason, result", ((False, REASON.quoted()), (True, f'"{REASON.reason}"')))
    def test_disabled(self, verbatim_reason, result):
        """It should disable Puppet, yield and enable Puppet on the hosts."""
        with self.puppet_hosts.disabled(REASON, verbatim_reason):
            self.mocked_remote_hosts.run_sync.assert_called_once_with(f"disable-puppet {result}")
            self.mocked_remote_hosts.run_sync.reset_mock()

        self.mocked_remote_hosts.run_sync.assert_called_once_with(f"enable-puppet {result}")

    def test_disabled_on_raise(self):
        """It should re-enable Puppet even if the yielded code raises exception.."""
        with pytest.raises(RuntimeError):
            with self.puppet_hosts.disabled(REASON):
                self.mocked_remote_hosts.run_sync.reset_mock()
                raise RuntimeError("Error")

        self.mocked_remote_hosts.run_sync.assert_called_once_with(f"enable-puppet {REASON.quoted()}")

    def test_check_enabled_ok(self):
        """It should check that all hosts have Puppet enabled."""
        host1 = nodeset("test1.example.com")
        host2 = nodeset("test2.example.com")
        results = [
            (host1, MsgTreeElem(b"0", parent=MsgTreeElem())),
            (host2, MsgTreeElem(b"0", parent=MsgTreeElem())),
        ]
        self.mocked_remote_hosts.run_sync.return_value = iter(results)
        self.puppet_hosts.check_enabled()

    def test_check_enabled_raise(self):
        """It should raise PuppetHostsCheckError if Puppet is disabled on some hosts."""
        host1 = nodeset("test1.example.com")
        host2 = nodeset("test2.example.com")
        results = [
            (host1, MsgTreeElem(b"0", parent=MsgTreeElem())),
            (host2, MsgTreeElem(b"1", parent=MsgTreeElem())),
        ]
        self.mocked_remote_hosts.run_sync.return_value = iter(results)
        with pytest.raises(
            puppet.PuppetHostsCheckError,
            match="Puppet is not enabled on those hosts: test2.example.com",
        ):
            self.puppet_hosts.check_enabled()

    def test_check_disabled_ok(self):
        """It should check that all hosts have Puppet disabled."""
        host1 = nodeset("test1.example.com")
        host2 = nodeset("test2.example.com")
        results = [
            (host1, MsgTreeElem(b"1", parent=MsgTreeElem())),
            (host2, MsgTreeElem(b"1", parent=MsgTreeElem())),
        ]
        self.mocked_remote_hosts.run_sync.return_value = iter(results)
        self.puppet_hosts.check_disabled()

    def test_check_disabled_raise(self):
        """It should raise PuppetHostsCheckError if Puppet is disabled on some hosts."""
        host1 = nodeset("test1.example.com")
        host2 = nodeset("test2.example.com")
        results = [
            (host1, MsgTreeElem(b"1", parent=MsgTreeElem())),
            (host2, MsgTreeElem(b"0", parent=MsgTreeElem())),
        ]
        self.mocked_remote_hosts.run_sync.return_value = iter(results)
        with pytest.raises(
            puppet.PuppetHostsCheckError,
            match="Puppet is not disabled on those hosts: test2.example.com",
        ):
            self.puppet_hosts.check_disabled()

    @pytest.mark.parametrize(
        "kwargs, expected",
        (
            ({}, ""),
            ({"enable_reason": REASON}, "--enable " + REASON.quoted()),
            ({"quiet": True}, "--quiet"),
            ({"failed_only": True}, "--failed-only"),
            ({"force": True}, "--force"),
            ({"attempts": 5}, "--attempts 5"),
        ),
    )
    def test_run_ok(self, kwargs, expected):
        """It should run Puppet with the specified arguments."""
        self.puppet_hosts.run(**kwargs)
        self.mocked_remote_hosts.run_sync.assert_called_once_with(
            puppet.Command(f"run-puppet-agent {expected}", timeout=300.0),
            batch_size=10,
        )

    def test_run_timeout(self):
        """It should run Puppet with the customized timeout."""
        self.puppet_hosts.run(timeout=30)
        self.mocked_remote_hosts.run_sync.assert_called_once_with(
            puppet.Command("run-puppet-agent ", timeout=30.0), batch_size=10
        )

    def test_first_run(self):
        """It should enable and Puppet with a very long timeout without using custom wrappers."""
        self.puppet_hosts.first_run()
        self.mocked_remote_hosts.run_sync.assert_called_once_with(
            "systemctl stop puppet.service",
            "systemctl reset-failed puppet.service || true",
            "puppet agent --enable",
            puppet.Command(
                "puppet agent --onetime --no-daemonize --verbose --no-splay --show_diff --no-usecacheonfailure",
                timeout=10800,
            ),
            print_output=False,
            print_progress_bars=False,
        )

    def test_first_run_not_systemd(self):
        """It should enable and Puppet with a very long timeout without using custom wrappers."""
        self.puppet_hosts.first_run(has_systemd=False)
        self.mocked_remote_hosts.run_sync.assert_called_once_with(
            "puppet agent --enable",
            puppet.Command(
                "puppet agent --onetime --no-daemonize --verbose --no-splay --show_diff --no-usecacheonfailure",
                timeout=10800,
            ),
            print_output=False,
            print_progress_bars=False,
        )

    def test_regenerate_certificate_ok(self):
        """It should delete and regenerate the Puppet certificate."""
        results = [
            (
                nodeset("test.example.com"),
                MsgTreeElem(PUPPET_GENERATE_CERTIFICATE_SUCCESS, parent=MsgTreeElem()),
            )
        ]
        self.mocked_remote_hosts.run_sync.side_effect = [iter(()), iter(results)]

        fingerprints = self.puppet_hosts.regenerate_certificate()

        self.mocked_remote_hosts.run_sync.assert_has_calls(
            [
                mock.call("rm -rfv /var/lib/puppet/ssl"),
                mock.call(puppet.Command("puppet agent --test --color=false", ok_codes=[]), print_output=False),
            ]
        )
        assert fingerprints == {"test.example.com": "00:FF"}

    def test_regenerate_certificate_raise(self):
        """It should raise PuppetHostsError if unable to find any of the fingerprint."""
        message = PUPPET_GENERATE_CERTIFICATE_SUCCESS.decode().replace(": 00:FF", "").encode()
        results = [(nodeset("test.example.com"), MsgTreeElem(message, parent=MsgTreeElem()))]
        self.mocked_remote_hosts.run_sync.side_effect = [iter(()), iter(results)]

        with pytest.raises(
            puppet.PuppetHostsError,
            match="Unable to find CSR fingerprints for all hosts",
        ):
            self.puppet_hosts.regenerate_certificate()

    def test_regenerate_certificate_errors(self):
        """It should raise PuppetHostsError and print the Puppet errors if unable to find any of the fingerprint."""
        results = [
            (
                nodeset("test.example.com"),
                MsgTreeElem(PUPPET_GENERATE_CERTIFICATE_FAILED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_remote_hosts.run_sync.side_effect = [iter(()), iter(results)]

        with pytest.raises(
            puppet.PuppetHostsError,
            match=(
                "test.example.com: Error: Could not request certificate: The certificate retrieved "
                "from the master does not match the agent's private key."
            ),
        ):
            self.puppet_hosts.regenerate_certificate()

    def test_wait_since_ok(self):
        """It should return immediately if there is already successful Puppet run since the given datetime."""
        last_run = datetime.utcnow()
        # timestamp() consider naive datetime objects as local time.
        last_run_string = str(int(last_run.replace(tzinfo=timezone.utc).timestamp()))
        start = last_run - timedelta(seconds=1)

        nodes = nodeset("test.example.com")
        results = [(nodes, MsgTreeElem(last_run_string.encode(), parent=MsgTreeElem()))]
        self.mocked_remote_hosts.run_sync.return_value = iter(results)
        self.mocked_remote_hosts.hosts = nodes

        self.puppet_hosts.wait_since(start)

        self.mocked_remote_hosts.run_sync.assert_called_once_with(
            (
                "source /usr/local/share/bash/puppet-common.sh && last_run_success && awk /last_run/'{ print $2 }' "
                '"${PUPPET_SUMMARY}"'
            ),
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

    # TODO: check why the following test_wait_since_* tests take longer (~4s each) when running the whole suite but are
    # quick if running only tests in this module (tox -e py34-unit -- -k test_puppet)
    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_since_timeout(self, mocked_sleep):
        """It should raise PuppetHostsCheckError if the successful Puppet run is too old within the timeout."""
        last_run = datetime.utcnow()
        # timestamp() consider naive datetime objects as local time.
        last_run_string = str(int(last_run.replace(tzinfo=timezone.utc).timestamp()))
        start = last_run + timedelta(seconds=1)

        nodes = nodeset("test.example.com")
        results = [(nodes, MsgTreeElem(last_run_string.encode(), parent=MsgTreeElem()))]
        self.mocked_remote_hosts.run_sync.side_effect = [results] * 60
        self.mocked_remote_hosts.hosts = nodes

        with pytest.raises(puppet.PuppetHostsCheckError, match="Successful Puppet run too old"):
            self.puppet_hosts.wait_since(start)

        assert mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_since_failed_execution(self, mocked_sleep):
        """It should raise PuppetHostsCheckError if fails to get the successful Puppet run within the timeout."""
        self.mocked_remote_hosts.run_sync.side_effect = RemoteExecutionError(1, "fail", iter(()))
        self.mocked_remote_hosts.hosts = nodeset("test.example.com")

        with pytest.raises(puppet.PuppetHostsCheckError, match="Unable to find a successful Puppet run"):
            self.puppet_hosts.wait_since(datetime.utcnow())

        assert mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_since_missing_host(self, mocked_sleep):
        """It should raise PuppetHostsCheckError unable to get the result from some host."""
        last_run = datetime.utcnow()
        # timestamp() consider naive datetime objects as local time.
        last_run_string = str(int(last_run.replace(tzinfo=timezone.utc).timestamp()))
        start = last_run - timedelta(seconds=1)

        nodes = nodeset("test[1-2].example.com")
        results = [
            (
                nodeset("test1.example.com"),
                MsgTreeElem(last_run_string.encode(), parent=MsgTreeElem()),
            )
        ]
        self.mocked_remote_hosts.run_sync.side_effect = [results] * 60
        self.mocked_remote_hosts.hosts = nodes

        with pytest.raises(
            puppet.PuppetHostsCheckError,
            match="Unable to get successful Puppet run from: test2.example.com",
        ):
            self.puppet_hosts.wait_since(start)

        assert mocked_sleep.called

    def test_wait_ok(self):
        """It should return immediately if there is already successful Puppet run since now."""
        last_run = datetime.utcnow() + timedelta(seconds=1)
        # timestamp() consider naive datetime objects as local time.
        last_run_string = str(int(last_run.replace(tzinfo=timezone.utc).timestamp()))

        nodes = nodeset("test.example.com")
        results = [(nodes, MsgTreeElem(last_run_string.encode(), parent=MsgTreeElem()))]
        self.mocked_remote_hosts.run_sync.return_value = iter(results)
        self.mocked_remote_hosts.hosts = nodes

        self.puppet_hosts.wait()

        self.mocked_remote_hosts.run_sync.assert_called_once_with(
            (
                "source /usr/local/share/bash/puppet-common.sh && last_run_success && awk /last_run/'{ print $2 }' "
                '"${PUPPET_SUMMARY}"'
            ),
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

    @pytest.mark.parametrize(
        "get_config_return, puppethost_side_effect",
        (
            (
                {"test1001.example.org": "true", "test1002.example.org": "true", "test1003.example.org": "false"},
                [
                    {"test1001.example.org": "eqiad.wmnet", "test1002.example.org": "codfw.wmnet"},
                    {"test1003.example.org": "dummy.puppetmast.er"},
                ],
            ),
            (
                {"test1001.example.org": "true", "test1002.example.org": "true", "test1003.example.org": "true"},
                [
                    {
                        "test1001.example.org": "eqiad.wmnet",
                        "test1002.example.org": "codfw.wmnet",
                        "test1003.example.org": "eqiad.wmnet",
                    },
                ],
            ),
            (
                {"test1001.example.org": "false", "test1002.example.org": "false", "test1003.example.org": "false"},
                [
                    {
                        "test1001.example.org": "dummy.puppetmast.er",
                        "test1002.example.org": "dummy.puppetmast.er",
                        "test1003.example.org": "dummy.puppetmast.er",
                    },
                ],
            ),
        ),
    )
    @mock.patch("spicerack.puppet.get_ca_via_srv_record")
    @mock.patch("spicerack.puppet.PuppetHosts")
    @mock.patch("spicerack.puppet.PuppetHosts.get_config")
    def test_get_ca_servers(
        self,
        mocked_get_config,
        mocked_puppet_hosts,
        mocked_get_ca_via_srv_record,
        get_config_return,
        puppethost_side_effect,
    ):
        """Test that get ca servers explodes multihost nodeset into single hosts."""
        expected_puppetmaster = "dummy.puppetmast.er"
        mocked_get_config.return_value = get_config_return
        mocked_puppet_hosts.return_value.get_config.side_effect = puppethost_side_effect
        mocked_get_ca_via_srv_record.return_value = expected_puppetmaster

        check_srv_domains = any(v == "true" for v in get_config_return)
        check_ca_server = any(v == "false" for v in get_config_return)

        result = self.puppet_hosts.get_ca_servers()

        mocked_get_config.assert_called_once_with("use_srv_records")
        if check_srv_domains:
            mocked_puppet_hosts.return_value.get_config.assert_any_call("srv_domain")
        if check_ca_server:
            mocked_puppet_hosts.return_value.get_config.assert_any_call("ca_server")

        assert "test1001.example.org" in result
        assert result["test1001.example.org"] == expected_puppetmaster
        assert "test1002.example.org" in result
        assert result["test1002.example.org"] == expected_puppetmaster
        assert "test1003.example.org" in result
        assert result["test1003.example.org"] == expected_puppetmaster

    def test_get_config_explodes_multihost_nodeset_into_single_hosts(self):
        """Test that get ca servers explodes multihost nodeset into single hosts."""
        expected_puppetmaster = "dummy.puppetmast.er"
        self.mocked_remote_hosts.run_sync.return_value = [
            (
                nodeset("test[0-1].example.com"),
                MsgTreeElem(expected_puppetmaster.encode(), parent=MsgTreeElem()),
            ),
        ]
        result = self.puppet_hosts.get_config("ca_server")
        self.mocked_remote_hosts.run_sync.assert_called_once()
        assert "test0.example.com" in result
        assert result["test0.example.com"] == expected_puppetmaster
        assert "test1.example.com" in result
        assert result["test1.example.com"] == expected_puppetmaster

    def test_get_config_handles_empty_result(self):
        """Test that get ca servers handles empty result."""
        self.mocked_remote_hosts.run_sync.return_value = []
        result = self.puppet_hosts.get_config("ca_server")
        self.mocked_remote_hosts.run_sync.assert_called_once()
        assert result == {}  # pylint: disable=use-implicit-booleaness-not-comparison

    def test_get_config_handles_multiple_results(self):
        """Test test get ca servers handles multiple results."""
        self.mocked_remote_hosts.run_sync.return_value = [
            (
                nodeset("test0.example.com"),
                MsgTreeElem(b"test0.puppetmast.er", parent=MsgTreeElem()),
            ),
            (
                nodeset("test1.example.com"),
                MsgTreeElem(b"test1.puppetmast.er", parent=MsgTreeElem()),
            ),
        ]
        result = self.puppet_hosts.get_config("ca_server")
        self.mocked_remote_hosts.run_sync.assert_called_once()
        assert "test0.example.com" in result
        assert result["test0.example.com"] == "test0.puppetmast.er"
        assert "test1.example.com" in result
        assert result["test1.example.com"] == "test1.puppetmast.er"

    def test_get_config_handles_multiple_lines_in_command_output(self):
        """Test test get ca servers handles multiple results."""
        self.mocked_remote_hosts.run_sync.return_value = [
            (
                nodeset("test0.example.com"),
                MsgTreeElem(
                    b"test0.puppetmast.er",
                    parent=MsgTreeElem(b"some message that should be ignored", parent=MsgTreeElem()),
                ),
            ),
        ]
        result = self.puppet_hosts.get_config("ca_server")
        self.mocked_remote_hosts.run_sync.assert_called_once()
        assert "test0.example.com" in result
        assert result["test0.example.com"] == "test0.puppetmast.er"


class TestPuppetServer:
    """Test class for the PuppetServer class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_server_host = mock.MagicMock(spec_set=RemoteHosts)
        self.mocked_server_host.__len__.return_value = 1
        self.puppet_server = puppet.PuppetServer(self.mocked_server_host)

    def test_init_raise(self):
        """It should raise PuppetServerError if the Puppet master host instance doesn't match one host."""
        self.mocked_server_host.__len__.return_value = 2
        with pytest.raises(
            puppet.PuppetServerError,
            match="The server_host instance must target only one host, got 2",
        ):
            puppet.PuppetServer(self.mocked_server_host)

    def test_server_host(self):
        """It should return the master host RemoteHosts instance."""
        assert self.puppet_server.server_host is self.mocked_server_host

    def test_delete(self):
        """It should delete the host from Puppet master and PuppetDB."""
        self.puppet_server.delete("test.example.com")
        self.mocked_server_host.run_sync.assert_called_once_with(
            "puppet node clean test.example.com", "puppet node deactivate test.example.com", print_progress_bars=False
        )

    def test_destroy_ok(self):
        """It should delete the certificate of the host in the Puppet CA."""
        results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(PUPPET_SERVER_CA_CERT_METADATA_SIGNED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.return_value = iter(results)

        self.puppet_server.destroy("test.example.com")
        self.mocked_server_host.run_sync.assert_called_with(
            "puppetserver ca clean --certname test.example.com", print_progress_bars=False
        )

    def test_destroy_missing(self):
        """It should not do anything if the certificate of the host in the Puppet CA is alredy missing."""
        results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(PUPPET_SERVER_CA_CERT_METADATA_MISSING, parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.return_value = iter(results)

        self.puppet_server.destroy("test.example.com")

        for call in self.mocked_server_host.mock_calls:
            if call.args and "ca clean" in call.args[0]:
                raise RuntimeError(f"Expected ca clean to not be called, got: {call}")

    def test_verify_ok(self):
        """It should verify that the host has a signed certificate in the Puppet CA."""
        results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(PUPPET_SERVER_CA_CERT_METADATA_SIGNED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.return_value = iter(results)

        assert self.puppet_server.verify("test.example.com") is None

        self.mocked_server_host.run_sync.assert_has_calls(
            [
                mock.call(
                    "puppetserver ca list --format json --certname test.example.com",
                    is_safe=True,
                    print_output=False,
                    print_progress_bars=False,
                ),
            ]
        )

    def test_verify_raise(self):
        """It should raise PuppetServerError if the certificate is not valid on the Puppet CA."""
        results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(PUPPET_SERVER_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.return_value = iter(results)

        with pytest.raises(
            puppet.PuppetServerError, match="Expected certificate for test.example.com to be signed, got: requested"
        ):
            self.puppet_server.verify("test.example.com")

        self.mocked_server_host.run_sync.assert_has_calls(
            [
                mock.call(
                    "puppetserver ca list --format json --certname test.example.com",
                    is_safe=True,
                    print_output=False,
                    print_progress_bars=False,
                ),
            ]
        )

    def test_sign_ok(self):
        """It should sign the certificate in the Puppet CA."""
        requested_results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(PUPPET_SERVER_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        signed_results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(PUPPET_SERVER_CA_CERT_METADATA_SIGNED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.side_effect = [
            iter(requested_results),
            iter(()),
            iter(signed_results),
        ]

        self.puppet_server.sign("test.example.com", "00:AA")

        self.mocked_server_host.run_sync.assert_has_calls(
            [
                mock.call(
                    "puppetserver ca list --format json --certname test.example.com",
                    is_safe=True,
                    print_output=False,
                    print_progress_bars=False,
                ),
                mock.call(
                    "puppetserver ca sign --certname test.example.com",
                    print_output=False,
                    print_progress_bars=False,
                ),
                mock.call(
                    "puppetserver ca list --format json --certname test.example.com",
                    is_safe=True,
                    print_output=False,
                    print_progress_bars=False,
                ),
            ]
        )

    def test_sign_wrong_state(self):
        """It should raise PuppetServerError if the certificate is not in the requested state."""
        results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(PUPPET_SERVER_CA_CERT_METADATA_SIGNED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.return_value = iter(results)

        with pytest.raises(
            puppet.PuppetServerError,
            match="Certificate for test.example.com not in requested state, got: signed",
        ):
            self.puppet_server.sign("test.example.com", "00:AA")

    def test_sign_wrong_fingerprint(self):
        """It should raise PuppetServerError if the fingerprint doesn't match the one in the CSR."""
        results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(PUPPET_SERVER_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.return_value = iter(results)

        with pytest.raises(
            puppet.PuppetServerError,
            match="CSR fingerprint 00:AA for test.example.com does not match provided fingerprint FF:FF",
        ):
            self.puppet_server.sign("test.example.com", "FF:FF")

    def test_sign_fail(self):
        """It should raise PuppetServerError if the sign operation fails."""
        requested_results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(PUPPET_SERVER_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        sign_results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(b"sign error", parent=MsgTreeElem()),
            )
        ]
        signed_results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(PUPPET_SERVER_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.side_effect = [
            iter(requested_results),
            iter(sign_results),
            iter(signed_results),
        ]

        with pytest.raises(
            puppet.PuppetServerError,
            match="Expected certificate for test.example.com to be signed, got: requested",
        ):
            self.puppet_server.sign("test.example.com", "00:AA")

    def test_wait_for_csr_already_ok(self):
        """It should return immediately if the certificate is already requested."""
        results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(PUPPET_SERVER_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.return_value = iter(results)

        self.puppet_server.wait_for_csr("test.example.com")

        self.mocked_server_host.run_sync.assert_called_once_with(
            "puppetserver ca list --format json --certname test.example.com",
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

    def test_wait_for_csr_fail(self):
        """It should raise PuppetServerError if the certificate is in a wrong state."""
        results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(PUPPET_SERVER_CA_CERT_METADATA_SIGNED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.return_value = iter(results)

        with pytest.raises(
            puppet.PuppetServerError,
            match="Expected certificate in requested state, got: signed",
        ):
            self.puppet_server.wait_for_csr("test.example.com")

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_for_csr_timeout(self, mocked_sleep):
        """It should raise PuppetServerCheckError if the certificate request doesn't appear."""
        results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(b"[]", parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.side_effect = [iter(results) for _ in range(10)]

        with pytest.raises(
            puppet.PuppetServerCheckError,
            match="No certificate found for hostname: test.example.com",
        ):
            self.puppet_server.wait_for_csr("test.example.com")

        assert mocked_sleep.called

    def test_get_certificate_metadata_ok(self):
        """It should return the metadata of the certificate for the host in the Puppet CA."""
        results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(PUPPET_SERVER_CA_CERT_METADATA_SIGNED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.return_value = iter(results)

        metadata = self.puppet_server.get_certificate_metadata("test.example.com")

        assert metadata == json.loads(PUPPET_SERVER_CA_CERT_METADATA_SIGNED.decode())["signed"][0]
        self.mocked_server_host.run_sync.assert_called_once_with(
            "puppetserver ca list --format json --certname test.example.com",
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

    @pytest.mark.parametrize(
        "json_output, exception, exception_message",
        (
            (
                b'{"signed":[{"name":"test.example.com"},{"name":"test.example.com"}]}',
                puppet.PuppetServerError,
                "Expected one result from Puppet CA, got 2",
            ),
            (
                b"[true]",
                puppet.PuppetServerCheckError,
                r"Expected a dict but got list: \[True\]",
            ),
            (
                b'{"missing":["test.example.com"]}',
                puppet.PuppetServerCheckError,
                r"The puppet server has no CSR for test.example.com",
            ),
            (
                b'{"signed":[{"name":"invalid.example.com"}]}',
                puppet.PuppetServerError,
                "Hostname mismatch invalid.example.com != test.example.com",
            ),
            (
                b'{"signed":[], "requested": []}',
                puppet.PuppetServerError,
                "Expected one result type from Puppet CA, got 2",
            ),
        ),
    )
    def test_get_certificate_metadata_raises(self, json_output, exception, exception_message):
        """It should raise PuppetServerError if the Puppet CA returns multiple certificates metadata."""
        results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(json_output, parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.return_value = iter(results)

        with pytest.raises(exception, match=exception_message):
            self.puppet_server.get_certificate_metadata("test.example.com")

    def test_run_json_command_none_zero_rc(self):
        """If the cumin commands returns 0 the function should still return data."""
        json_output = b'{"missing":["test.example.com"]}'
        results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(json_output, parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.return_value = iter(results)
        side_effect = RemoteExecutionError(1, "Cumin, execution failed", iter(results))
        self.mocked_server_host.run_sync.side_effect = side_effect
        assert self.puppet_server._run_json_command("") == json.loads(json_output)  # pylint: disable=protected-access

    def test_hiera_lookup(self):
        """It should return the metadata of the certificate for the host in the Puppet CA."""
        results = [
            (
                nodeset("puppetserver.example.com"),
                MsgTreeElem(b"misc", parent=MsgTreeElem()),
            )
        ]
        self.mocked_server_host.run_sync.return_value = iter(results)

        assert self.puppet_server.hiera_lookup("test.example.com", "cluster") == "misc"

        self.mocked_server_host.run_sync.assert_called_once_with(
            "puppet lookup --render-as s --compile --node test.example.com cluster 2>/dev/null",
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )


class TestPuppetMaster:
    """Test class for the PuppetMaster class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_master_host = mock.MagicMock(spec_set=RemoteHosts)
        self.mocked_master_host.__len__.return_value = 1
        self.puppet_master = puppet.PuppetMaster(self.mocked_master_host)

    def test_master_host(self):
        """It should return the master host RemoteHosts instance."""
        assert self.puppet_master.master_host is self.mocked_master_host

    def test_destroy(self):
        """It should delete the certificate of the host in the Puppet CA."""
        self.puppet_master.destroy("test.example.com")
        self.mocked_master_host.run_sync.assert_called_once_with(
            "puppet ca --disable_warnings deprecations destroy test.example.com", print_progress_bars=False
        )

    def test_verify_ok(self):
        """It should verify that the host has a signed certificate in the Puppet CA."""
        json_output = b'{"host":"test.example.com","valid":true}'
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(json_output, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        self.puppet_master.verify("test.example.com")
        self.mocked_master_host.run_sync.assert_called_once_with(
            "puppet ca --disable_warnings deprecations --render-as json verify test.example.com",
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

    def test_verify_raise(self):
        """It should raise PuppetServerError if the certificate is not valid on the Puppet CA."""
        json_output = b'{"host":"test.example.com","valid":false,"error":"Error message"}'
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(json_output, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        with pytest.raises(
            puppet.PuppetServerError,
            match="Invalid certificate for test.example.com: Error message",
        ):
            self.puppet_master.verify("test.example.com")

        self.mocked_master_host.run_sync.assert_called_once_with(
            "puppet ca --disable_warnings deprecations --render-as json verify test.example.com",
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

    def test_verify_no_output(self):
        """It should raise PuppetServerError if there is no output from the executed command."""
        self.mocked_master_host.run_sync.return_value = iter(())
        with pytest.raises(
            puppet.PuppetServerError,
            match=(
                re.escape(
                    "Got no output from Puppet server while executing command (rc: 0): "
                    "puppet ca --disable_warnings deprecations --render-as json verify test.example.com"
                )
            ),
        ):
            self.puppet_master.verify("test.example.com")

    def test_verify_invalid_json(self):
        """It should raise PuppetServerError if there output of the executed command is invalid JSON."""
        json_output = b'{"host":"test.example.com",,}'
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(json_output, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        with pytest.raises(
            puppet.PuppetServerError,
            match=(
                re.escape(
                    "Unable to parse Puppet server response for command (rc: 0): "
                    '"puppet ca --disable_warnings deprecations --render-as json verify test.example.com": '
                    r'{"host":"test.example.com",,}'
                )
            ),
        ):
            self.puppet_master.verify("test.example.com")

    def test_sign_ok(self):
        """It should sign the certificate in the Puppet CA."""
        requested_results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        signed_results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_SIGNED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.side_effect = [
            iter(requested_results),
            iter(()),
            iter(signed_results),
        ]

        self.puppet_master.sign("test.example.com", "00:AA")

        self.mocked_master_host.run_sync.assert_has_calls(
            [
                mock.call(
                    "puppet ca --disable_warnings deprecations --render-as json "
                    r'list --all --subject "^test\.example\.com$"',
                    is_safe=True,
                    print_output=False,
                    print_progress_bars=False,
                ),
                mock.call(
                    "puppet cert --disable_warnings deprecations sign test.example.com",
                    print_output=False,
                    print_progress_bars=False,
                ),
                mock.call(
                    "puppet ca --disable_warnings deprecations --render-as json "
                    r'list --all --subject "^test\.example\.com$"',
                    is_safe=True,
                    print_output=False,
                    print_progress_bars=False,
                ),
            ]
        )

    def test_sign_wrong_state(self):
        """It should raise PuppetServerError if the certificate is not in the requested state."""
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_SIGNED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        with pytest.raises(
            puppet.PuppetServerError,
            match="Certificate for test.example.com not in requested state, got: signed",
        ):
            self.puppet_master.sign("test.example.com", "00:AA")

    def test_sign_wrong_fingerprint(self):
        """It should raise PuppetServerError if the fingerprint doesn't match the one in the CSR."""
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        with pytest.raises(
            puppet.PuppetServerError,
            match="CSR fingerprint 00:AA for test.example.com does not match provided fingerprint FF:FF",
        ):
            self.puppet_master.sign("test.example.com", "FF:FF")

    def test_sign_fail(self):
        """It should raise PuppetServerError if the sign operation fails."""
        requested_results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        sign_results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(b"sign error", parent=MsgTreeElem()),
            )
        ]
        signed_results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_REQUESTED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.side_effect = [
            iter(requested_results),
            iter(sign_results),
            iter(signed_results),
        ]

        with pytest.raises(
            puppet.PuppetServerError,
            match="Expected certificate for test.example.com to be signed, got: requested",
        ):
            self.puppet_master.sign("test.example.com", "00:AA")

    def test_get_certificate_metadata_ok(self):
        """It should return the metadata of the certificate for the host in the Puppet CA."""
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(PUPPET_CA_CERT_METADATA_SIGNED, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        metadata = self.puppet_master.get_certificate_metadata("test.example.com")

        assert metadata == json.loads(PUPPET_CA_CERT_METADATA_SIGNED.decode())[0]
        self.mocked_master_host.run_sync.assert_called_once_with(
            "puppet ca --disable_warnings deprecations --render-as json list "
            r'--all --subject "^test\.example\.com$"',
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

    @pytest.mark.parametrize(
        "json_output, exception, exception_message",
        (
            (
                b"{}",
                puppet.PuppetServerCheckError,
                "No certificate found for hostname: test.example.com",
            ),
            (
                b'[{"name":"test.example.com"},{"name":"test.example.com"}]',
                puppet.PuppetServerError,
                "Expected one result from Puppet CA, got 2",
            ),
            (
                b'[{"name":"invalid.example.com"}]',
                puppet.PuppetServerError,
                "Hostname mismatch invalid.example.com != test.example.com",
            ),
        ),
    )
    def test_get_certificate_metadata_raises(self, json_output, exception, exception_message):
        """It should raise PuppetServerError if the Puppet CA returns multiple certificates metadata."""
        results = [
            (
                nodeset("puppetmaster.example.com"),
                MsgTreeElem(json_output, parent=MsgTreeElem()),
            )
        ]
        self.mocked_master_host.run_sync.return_value = iter(results)

        with pytest.raises(exception, match=exception_message):
            self.puppet_master.get_certificate_metadata("test.example.com")



"""Dnsdisc module tests."""
import logging
from collections import namedtuple
from ipaddress import ip_address
from unittest import mock

import dns
import pytest
from dns.exception import DNSException
from wmflib.config import load_yaml_config

from spicerack.dnsdisc import Discovery, DiscoveryCheckError, DiscoveryError
from spicerack.tests import get_fixture_path

MockedRecord = namedtuple("Record", ["address"])
DISCOVERY_RESPONSE_OK = """id 57640
opcode QUERY
rcode NOERROR
flags QR RD RA
;QUESTION
record1.discovery.wmnet. IN A
;ANSWER
record1.discovery.wmnet. 300 IN A 10.10.10.10
;AUTHORITY
;ADDITIONAL
"""
DISCOVERY_RESPONSE_ERROR = """id 6890
opcode QUERY
rcode NXDOMAIN
flags QR RD RA
;QUESTION
record1.discovery.wmnet. IN A
;ANSWER
;AUTHORITY
wmnet. 2434 IN SOA ns0.wikimedia.org. hostmaster.wikimedia.org. 2023031514 43200 7200 1209600 3600
;ADDITIONAL
"""


def mock_obj(datacenter, svc, pooled):
    """Creates a mock conftool object."""
    obj = mock.MagicMock()
    obj.tags = {"dnsdisc": svc}
    obj.name = datacenter
    obj.pooled = pooled
    return obj


def get_mocked_dns_query_message(fail: bool) -> dns.message.Message:
    """Get a mocked low level response to be used with udp_with_fallback() mocks."""
    response = DISCOVERY_RESPONSE_ERROR if fail else DISCOVERY_RESPONSE_OK
    message = dns.message.from_text(response)
    message.find_rrset(
        message.answer,
        dns.name.from_text("record1.discovery.wmnet"),
        dns.rdataclass.from_text("IN"),
        dns.rdatatype.from_text("A"),
        create=True,
    )
    return message


class MockedQuery:
    """Class to mock a return object from a call to dns.resolver.query()."""

    def __init__(self, record):
        """Initialize it with the query."""
        if record == "fail.svc.eqiad.wmnet":
            self.address = "10.1.1.1"
            self.ttl = 600
        elif record.startswith("raise"):
            raise DNSException
        else:
            self.address = "10.0.0.1"
            self.ttl = 10

        self.record = MockedRecord(self.address)

    def __getitem__(self, _):
        """Allow indexing access."""
        return self.record

    def __iter__(self):
        """Allow iterating on results."""
        return iter([self.record])


class MockDnsResult:
    """Mocks dns.message.ChainingResult."""

    def __init__(self, ipaddr: str):
        """Creates the chaining result."""
        # Mocks a rrset
        a_record = mock.MagicMock()
        a_record.address = ipaddr
        # Where dns.resolver.Answer gets the data from
        self.answer = [a_record]
        self.canonical_name = "a_name"
        self.minimum_ttl = 10


class TestDiscovery:
    """Discovery class tests."""

    @mock.patch("spicerack.dnsdisc.resolver", autospec=True)
    def setup_method(self, _, mocked_resolver):
        """Initialize the test environment for Discovery."""
        # pylint: disable=attribute-defined-outside-init
        self.records = ["record1", "record2"]
        self.conftool_records = "(" + "|".join(self.records) + ")"

        self.mocked_confctl = mock.MagicMock()
        self.authdns_servers = load_yaml_config(get_fixture_path("discovery", "authdns.yaml"))
        self.mocked_resolver = mocked_resolver
        self.mocked_resolver.Resolver.return_value.query = MockedQuery

        self.discovery = Discovery(
            conftool=self.mocked_confctl, authdns_servers=self.authdns_servers, records=self.records, dry_run=False
        )
        self.discovery_single = Discovery(
            conftool=self.mocked_confctl, authdns_servers=self.authdns_servers, records=self.records[0:1], dry_run=False
        )
        self.discovery_dry_run = Discovery(
            conftool=self.mocked_confctl, authdns_servers=self.authdns_servers, records=self.records
        )

    def test_resolve_address(self):
        """Calling resolve_address() sould raise DiscoveryError if unable to resolve the address."""
        with pytest.raises(DiscoveryError, match="Unable to resolve raise.svc.eqiad.wmnet"):
            self.discovery_single.resolve_address("raise.svc.eqiad.wmnet")

    def test_update_ttl(self):
        """Calling update_ttl() should update the TTL of the conftool objects."""
        self.discovery.update_ttl(10)
        self.mocked_confctl.assert_has_calls([mock.call.set_and_verify("ttl", 10, dnsdisc=self.conftool_records)])

    @mock.patch("wmflib.decorators.time.sleep")
    def test_update_ttl_dry_run(self, mocked_sleep):
        """Calling update_ttl() in DRY-RUN mode should not raise when verifying the TTL."""
        self.discovery_dry_run.update_ttl(20)
        assert not mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep")
    def test_update_ttl_ko(self, mocked_sleep):
        """Calling update_ttl() should raise DiscoveryCheckError if unable to verify the value."""
        with pytest.raises(DiscoveryCheckError, match="Expected TTL '20', got '10'"):
            self.discovery.update_ttl(20)

        assert mocked_sleep.called

    def test_check_ttl_ok(self):
        """Calling check_ttl() should verify that the correct TTL is returned by the authoritative nameservers."""
        self.discovery.check_ttl(10)

    @pytest.mark.parametrize(
        "expected, name",
        (
            ("record1.discovery.wmnet TTL is correct", "discovery_single"),
            ("['record1', 'record2'] discovery.wmnet TTLs are correct", "discovery"),
        ),
    )
    def test_check_ttl_log_message(self, expected, name, caplog):
        """The log message of check_ttl should reflect if a singular or multiple records matches."""
        with caplog.at_level(logging.INFO):
            getattr(self, name).check_ttl(10)
        assert expected in caplog.text

    @mock.patch("wmflib.decorators.time.sleep")
    def test_check_ttl_ko(self, mocked_sleep):
        """Calling check_ttl() should raise DiscoveryCheckError if the check fails."""
        with pytest.raises(DiscoveryCheckError, match="Expected TTL '20', got '10'"):
            self.discovery.check_ttl(20)

        assert mocked_sleep.called

    def test_check_record_ok(self):
        """Calling check_record() should verify that a record has a certain value on the nameservers."""
        self.discovery.check_record(self.records[0], "ok.svc.eqiad.wmnet")

    @mock.patch("wmflib.decorators.time.sleep")
    def test_check_record_ko(self, mocked_sleep):
        """Calling check_record() should raise DiscoveryError if unable to check the records."""
        with pytest.raises(
            DiscoveryError,
            match=f"Resolved record {self.records[0]} with the wrong IP",
        ):
            self.discovery.check_record(self.records[0], "fail.svc.eqiad.wmnet")

        assert mocked_sleep.called

    def test_resolve(self):
        """Calling resolve() should raise a DiscoveryError if unable to resolve the address."""
        with pytest.raises(DiscoveryError, match="Unable to resolve raise.discovery.wmnet from authdns"):
            for _ in self.discovery.resolve(name="raise"):
                pass

    @pytest.mark.parametrize("func, value", (("pool", True), ("depool", False)))
    def test_pool(self, func, value):
        """Calling pool() should update the pooled value of the conftool objects to True."""
        getattr(self.discovery, func)("eqiad")
        self.mocked_confctl.set_and_verify.assert_called_once_with(
            "pooled", value, dnsdisc=self.conftool_records, name="eqiad"
        )

    def test_active_datacenters(self):
        """The list of active datacenter is correctly composed."""
        self.mocked_confctl.get.return_value = [
            mock_obj("dcA", "svcA", True),
            mock_obj("dcB", "svcA", False),
            mock_obj("dcA", "svcB", True),
            mock_obj("dcB", "svcB", True),
        ]
        expected = {"svcA": ["dcA"], "svcB": ["dcA", "dcB"]}
        assert expected == self.discovery.active_datacenters
        self.mocked_confctl.get.assert_called_once_with(dnsdisc=self.conftool_records)

    def test_check_if_depoolable_ok(self):
        """No exception is raised if the active datacenters are ok."""
        self.mocked_confctl.get.return_value = [
            mock_obj("dcA", "svcA", False),
            mock_obj("dcB", "svcA", True),
            mock_obj("dcA", "svcB", True),
            mock_obj("dcB", "svcB", True),
        ]
        self.discovery.check_if_depoolable("dcA")
        # now let's assume a service is competely down. We should still get
        # a green light.
        self.mocked_confctl.get.return_value = [
            mock_obj("dcA", "svcA", False),
            mock_obj("dcB", "svcA", False),
            mock_obj("dcA", "svcB", True),
            mock_obj("dcB", "svcB", True),
        ]
        self.discovery.check_if_depoolable("dcA")

    def test_check_if_depoolable_ko(self):
        """A DiscoveryError is raised when a service would be taken out of commission."""
        self.mocked_confctl.get.return_value = [
            mock_obj("dcA", "svcA", False),
            mock_obj("dcB", "svcA", True),
            mock_obj("dcA", "svcB", False),
            mock_obj("dcB", "svcB", False),
        ]
        with pytest.raises(
            DiscoveryError,
            match="Services svcA cannot be depooled as they are only active in dcB",
        ):
            self.discovery.check_if_depoolable("dcB")
        self.mocked_confctl.get.return_value = [
            mock_obj("dcA", "svcA", False),
            mock_obj("dcB", "svcA", True),
            mock_obj("dcA", "svcB", False),
            mock_obj("dcB", "svcB", True),
        ]
        with pytest.raises(
            DiscoveryError,
            match="Services svcA, svcB cannot be depooled as they are only active in dcB",
        ):
            self.discovery.check_if_depoolable("dcB")

    def test_check_if_depoolable_ko_dry_run(self, caplog):
        """Doesn't raise exception when a service would be taken out of commission but in dry-run mode."""
        self.mocked_confctl.get.return_value = [
            mock_obj("dcA", "svcA", False),
            mock_obj("dcB", "svcA", True),
            mock_obj("dcA", "svcB", False),
            mock_obj("dcB", "svcB", False),
        ]
        with caplog.at_level(logging.DEBUG):
            self.discovery_dry_run.check_if_depoolable("dcB")
        assert "cannot be depooled as they are only active in" in caplog.text

    def test_resolve_with_client_ip(self):
        """Correctly returns the ip address for the query."""
        with mock.patch("dns.query.udp_with_fallback") as mock_query:
            retval = {name: ip_address("10.10.10.10") for name in self.authdns_servers}
            mock_query.return_value = (get_mocked_dns_query_message(fail=False), False)
            assert self.discovery_single.resolve_with_client_ip("record1", ip_address("10.24.1.0")) == retval

    def test_resolve_bad_record(self):
        """Requesting a record that is not present will raise a DiscoveryError."""
        with pytest.raises(DiscoveryError, match="Record 'nope' not found"):
            self.discovery_single.resolve_with_client_ip("nope", ip_address("10.24.1.0"))

    def test_resolve_bad_query(self):
        """Any exception thrown by dns.query.udp_with_fallback will result in a DiscoveryError."""
        with mock.patch("dns.query.udp_with_fallback") as mock_query:
            mock_query.side_effect = ValueError("test this")
            with pytest.raises(DiscoveryError, match="Unable to resolve record1.discovery.wmnet"):
                self.discovery_single.resolve_with_client_ip("record1", ip_address("10.24.1.0"))

    def test_resolve_bad_answer(self):
        """When the query response is empty or contains an invalid IP, a DiscoveryError is raised."""
        with mock.patch("dns.query.udp_with_fallback") as mock_query:
            mock_query.return_value = (get_mocked_dns_query_message(fail=True), False)
            with pytest.raises(DiscoveryError, match="Unable to resolve record1.discovery.wmnet"):
                self.discovery_single.resolve_with_client_ip("record1", ip_address("10.24.1.0"))



"""Sphinx configuration checker.

- Check that all the existing Spicerack modules are listed in the API index documentation.
- Check that all the modules listed in the API index documentation exists in Spicerack.
- Check that all the listed modules have its own file in the api/ documentation directory.
  Sphinx would raise a warning but not fail in this case.
"""
import argparse
import sys
from pathlib import Path
from pkgutil import iter_modules

from setuptools import find_packages

import spicerack

DOC_API_BASE_PATH = Path("doc/source/api")
DOC_API_INDEX_PATH = DOC_API_BASE_PATH / "index.rst"


def main(base_path):
    """Perform the check."""
    base_path = Path(spicerack.__path__[0]) / ".."
    spicerack_modules = set()
    for package in find_packages(base_path):
        if package.startswith("spicerack.tests"):  # Skip all tests.
            continue

        if package != "spicerack":  # Do not include the main spicerack package.
            spicerack_modules.add(package)
        package_path = base_path / package.replace(".", "/")
        for module_info in iter_modules([str(package_path)]):
            if not module_info.ispkg and not module_info.name.startswith("_"):
                spicerack_modules.add(f"{package}.{module_info.name}")

    with open(DOC_API_INDEX_PATH) as f:
        api_index_lines = f.readlines()

    doc_api_modules = {line.strip() for line in api_index_lines if line.strip().startswith("spicerack.")}

    ret = 0
    if spicerack_modules - doc_api_modules:
        print(f"Spicerack modules that are not listed in {DOC_API_INDEX_PATH}: {spicerack_modules - doc_api_modules}")
        ret += 1
    if doc_api_modules - spicerack_modules:
        print(
            f"Documented modules in {DOC_API_INDEX_PATH} that are missing in Spicerack: "
            f"{doc_api_modules - spicerack_modules}"
        )
        ret += 1

    doc_api_files = [f"{name}.rst" for name in doc_api_modules]
    missing_doc_api_files = [file_name for file_name in doc_api_files if not (DOC_API_BASE_PATH / file_name).is_file()]
    if missing_doc_api_files:
        print(f"Missing documentation files in {DOC_API_BASE_PATH}: {missing_doc_api_files}")
        ret += 1

    if ret == 0:
        print("All Spicerack modules are documented")

    return ret


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Check that all Spicerack modules are documented")
    parser.add_argument("base_path", help="Path to the root of the spicerack repository")
    args = parser.parse_args()

    sys.exit(main(args.base_path))



"""Puppet module tests."""
from unittest import mock

import pytest

from spicerack import apt
from spicerack.remote import RemoteHosts

APT_GET_BASE_COMMAND = (
    'DEBIAN_FRONTEND=noninteractive /usr/bin/apt-get --quiet --yes --option Dpkg::Options::="--force-confdef" '
    '--option Dpkg::Options::="--force-confold"'
)


class TestAptGetHosts:
    """Test class for the AptGetHosts class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_remote_hosts = mock.MagicMock(spec_set=RemoteHosts)
        self.mocked_remote_hosts.__len__.return_value = 1
        self.apt_get_hosts = apt.AptGetHosts(self.mocked_remote_hosts)

    @pytest.mark.parametrize(
        "kwargs",
        (
            {},
            {"batch_size": 3},
        ),
    )
    def test_update(self, kwargs):
        """It should run apt-get update on the hosts with the given parameters."""
        self.apt_get_hosts.update(**kwargs)
        self.mocked_remote_hosts.run_sync.assert_has_calls([mock.call(f"{APT_GET_BASE_COMMAND} update", **kwargs)])

    @pytest.mark.parametrize(
        "packages, kwargs",
        (
            (["package1"], {}),
            (["package1", "package2"], {}),
            (["package1"], {"batch_size": 3}),
        ),
    )
    def test_install(self, packages, kwargs):
        """It should run apt-get install on the hosts for the given packages."""
        self.apt_get_hosts.install(*packages, **kwargs)
        pkgs = " ".join(packages)
        self.mocked_remote_hosts.run_sync.assert_has_calls(
            [mock.call(f"{APT_GET_BASE_COMMAND} install {pkgs}", **kwargs)]
        )

    def test_install_no_packages(self):
        """It should raise an AptGetError if there were no packages provided."""
        with pytest.raises(apt.AptGetError, match=r"No packages to install were provided\."):
            self.apt_get_hosts.install()



"""Top level cookbook."""
import argparse

__title__ = __doc__
MAX_CONCURRENCY = 5
LOCK_TTL = 120


def argument_parser():
    """As required by spicerack._cookbook."""
    return argparse.ArgumentParser("Argparse")


def run(args, _):
    """As required by spicerack._cookbook."""
    print(args)
    return 0



"""Spicerack extender for the external modules."""
from spicerack_ext.cool_feature import CoolFeature

from spicerack import SpicerackExtenderBase


class SpicerackExtender(SpicerackExtenderBase):
    """Spicerack extender to add custom accessors."""

    def cool_feature(self, feature: str) -> CoolFeature:
        """Get a CoolFeature instance.

        Arguments:
            feature: the feature to make cool.

        Returns:
            spicerack_ext.cool_feature.CoolFeature: the cool feature instance.

        """
        return CoolFeature(feature, dry_run=self._spicerack.dry_run)


class SpicerackBadExtender:
    """An extender that doesn't inherit from SpicerackExtenderBase."""



"""k8s module tests."""
from http import HTTPStatus
from unittest import mock

import kubernetes
import pytest
from kubernetes.client.models import V1Taint

from spicerack import k8s


def mock_list_value(val):
    """Helper for populate_mock."""
    if isinstance(val, dict):
        m = mock.MagicMock()
        populate_mock(m, val)
        return m
    # For non-objects we can just return the value.
    return val


def populate_mock(mocked, attrs):
    """Populates a mock with attributes and functions."""
    for name, value in attrs.items():
        target = getattr(mocked, name)
        if isinstance(value, dict):
            populate_mock(target, value)
        elif isinstance(value, list):
            setattr(mocked, name, [mock_list_value(v) for v in value])
        else:
            setattr(mocked, name, value)


class KubeTestBase:
    """Base class for testing all subclasses."""

    node_test_cases = {
        "schedulable": {"metadata": {"name": "node1"}, "spec": {"unschedulable": None, "taints": None}},
        "unschedulable": {"metadata": {"name": "node2"}, "spec": {"unschedulable": True, "taints": None}},
        "tainted": {
            "metadata": {"name": "node3"},
            "spec": {"unschedulable": None, "taints": [V1Taint(effect="NoExecute", key="dedicated", value="kask")]},
        },
    }
    pod_test_cases = {
        "orphaned": {
            "metadata": {"owner_references": None, "name": "foo", "namespace": "bar", "annotations": ["sigh"]},
            "status": {"phase": "Running"},
            "spec": {"termination_grace_period_seconds": 30},
        },
        "replicaset": {
            "metadata": {"owner_references": [{"kind": "ReplicaSet"}], "name": "foo", "namespace": "bar"},
            "status": {"phase": "Running"},
            "spec": {"termination_grace_period_seconds": 1},
        },
        "daemonset": {
            "metadata": {
                "owner_references": [{"kind": "DaemonSet"}, {"kind": "ReplicaSet"}],
                "name": "calico",
                "namespace": "bar",
            },
            "status": {"phase": "Running"},
            "spec": {"termination_grace_period_seconds": 30},
        },
        "finished": {
            "metadata": {
                "owner_references": [{"kind": "ReplicaSet"}],
                "name": "foo",
                "namespace": "bar",
            },
            "status": {"phase": "Succeeded"},
            "spec": {"termination_grace_period_seconds": 300},
        },
        "mirror": {
            "metadata": {
                "owner_references": [{"kind": "ReplicaSet"}],
                "name": "foo",
                "namespace": "bar",
                "annotations": ["kubernetes.io/config.mirror"],
            },
            "status": {"phase": "Running"},
        },
        "invalid_ref": {
            "metadata": {
                "owner_references": [],  # this should never happen, but better safe than sorry.
                "name": "foo",
                "namespace": "bar",
            },
            "status": {"phase": "Running"},
            "spec": {"termination_grace_period_seconds": 30},
        },
    }

    def setup_method(self):
        """Setup a mock api to work with."""
        # pylint: disable=attribute-defined-outside-init
        self._api = mock.MagicMock(spec=k8s.KubernetesApiFactory)
        self._coreapi = self._api.core.return_value

    def node_from_test_case(self, label):
        """Get an api node object from a test label."""
        return self._get_from_test_case("node", label)

    def pod_from_test_case(self, label):
        """Get an api pod object from a test label."""
        return self._get_from_test_case("pod", label)

    def _get_from_test_case(self, what, label):
        if what == "pod":
            cases = self.pod_test_cases
            cls = kubernetes.client.models.v1_pod.V1Pod
        elif what == "node":
            cases = self.node_test_cases
            cls = kubernetes.client.models.v1_node.V1Node
        else:
            raise ValueError(f"no test cases for '{what}'")

        try:
            attrs = cases[label]
        except KeyError:
            return None
        obj = mock.MagicMock(spec=cls)
        populate_mock(obj, attrs)
        return obj


class TestKubernetes(KubeTestBase):
    """Test the kubernetes object that is wired into spicerack eventually."""

    @pytest.fixture
    def kube(self):
        """Fixture to get a test object."""
        k = k8s.Kubernetes("group", "cluster", dry_run=False)
        k.api = self._api
        return k

    def test_initialize_api(self):
        """Values are passed to the api correctly."""
        k = k8s.Kubernetes("group", "cluster")
        assert k.api.cluster == "cluster"

    def test_initialize_dry_run(self, kube):
        """Dry run is initialized correctly."""
        assert kube.dry_run is False
        assert k8s.Kubernetes("group", "cluster", dry_run=True).dry_run is True

    def test_get_node(self, kube):
        """Getting a node works."""
        list_node = mock.MagicMock()
        list_node.items = [self.node_from_test_case("schedulable")]
        self._coreapi.list_node.return_value = list_node
        node = kube.get_node("node1")
        assert node.name == "node1"
        self._coreapi.list_node.assert_called_with(field_selector="metadata.name=node1")

    def test_get_pod(self, kube):
        """Getting a pod works."""
        with mock.patch("spicerack.k8s.KubernetesNode._get") as mocked:
            mocked.return_value = self.pod_from_test_case("replicaset")
            pod = kube.get_pod("bar", "foo")
            assert pod.name == "foo"
            assert pod.is_terminated() is False


class TestKubernetesApiFactory:
    """Test the api factory class."""

    def setup_method(self):
        """Setup a mock api to work with."""
        # pylint: disable=attribute-defined-outside-init
        self._api = k8s.KubernetesApiFactory("cluster")

    @mock.patch("kubernetes.config.load_kube_config")
    def test_configuration(self, mocked):
        """The configuration is correctly loaded."""
        my_config = kubernetes.client.Configuration()

        with mock.patch("kubernetes.client.Configuration") as conf:
            conf.return_value = my_config
            assert self._api.configuration("cpt_harlock") == my_config
            conf.assert_called()
            mocked.assert_called_with(
                config_file="/etc/kubernetes/cpt_harlock-cluster.config", client_configuration=my_config
            )

    @mock.patch("kubernetes.config.load_kube_config")
    def test_config_cached(self, mocked):
        """A configuration gets only loaded once."""
        self._api.configuration("cpt_harlock")
        self._api.configuration("cpt_harlock")
        assert mocked.call_count == 1

    def test_config_invalid_user(self):
        """An invalid user will raise an exception."""
        with pytest.raises(k8s.KubernetesError):
            self._api.configuration("../../etc/passwd")

    @mock.patch("kubernetes.config.load_kube_config")
    def test_config_bad_config(self, mocked):
        """An invalid or inexistent file will cause an exception."""
        mocked.side_effect = kubernetes.config.config_exception.ConfigException("Arcadia!")
        with pytest.raises(k8s.KubernetesError):
            self._api.configuration("cpt_harlock")

    def test_core(self):
        """Get the core api interface."""
        self._api.configuration = mock.MagicMock(return_value=kubernetes.client.Configuration())
        assert isinstance(self._api.core(), kubernetes.client.CoreV1Api)

    def test_batch(self):
        """Get the batch api interface."""
        self._api.configuration = mock.MagicMock(return_value=kubernetes.client.Configuration())
        assert isinstance(self._api.batch(), kubernetes.client.BatchV1Api)


class TestKubernetesNode(KubeTestBase):
    """Test the KubernetesNode class."""

    @pytest.fixture
    def node(self, label):
        """Returns a node from the test cases."""
        _node = self.node_from_test_case(label)
        if _node is None:
            fqdn = None
        else:
            fqdn = _node.metadata.name
        return k8s.KubernetesNode(fqdn, self._api, dry_run=False, init_obj=_node)

    @pytest.fixture
    def dry_run(self, label):
        """Returns a node, dry-run set to true."""
        _n = self.node_from_test_case(label)
        self._coreapi.list_node.return_value = self.list_nodes(_n.metadata.name)
        k = k8s.Kubernetes("group", "cluster", dry_run=True)
        k.api = self._api
        return k.get_node(_n.metadata.name)

    def list_nodes(self, name):
        """Returns a node list."""
        items = []
        for label in self.node_test_cases:
            n = self.node_from_test_case(label)
            n.metadata.annotations = ["refreshed"]
            if n.metadata.name == name:
                items.append(n)
        m = mock.MagicMock()
        m.items = items
        return m

    def test_bad_init(self):
        """A bad initial object will cause an exception."""
        n = self.node_from_test_case("schedulable")
        with pytest.raises(k8s.KubernetesError):
            k8s.KubernetesNode("fqdn", self._api, init_obj=n)

    @pytest.mark.parametrize("label,expected", [("schedulable", True), ("unschedulable", False)])
    def test_is_schedulable(self, node, expected):
        """Is the node schedulable or not."""
        assert node.is_schedulable() is expected

    @pytest.mark.parametrize("label,expected", [("schedulable", "node1"), ("unschedulable", "node2")])
    def test_name(self, node, expected):
        """Test fetching the node name."""
        assert node.name == expected

    @pytest.mark.parametrize("label,has_calls", [("schedulable", True), ("unschedulable", False)])
    def test_cordon(self, node, has_calls):
        """Cordoning a schedulable node works."""
        patch_expected = self.node_from_test_case("schedulable")
        patch_expected.spec.unschedulable = True
        self._coreapi.patch_node.return_value = patch_expected
        # Happy path, will not raise exception.
        node.cordon()
        if has_calls:
            # Check the call to the api is what we expect.
            self._coreapi.patch_node.assert_called_with("node1", {"spec": {"unschedulable": True}})
        else:
            self._coreapi.patch_node.assert_not_called()
        assert node.is_schedulable() is False

    @pytest.mark.parametrize("label", ["schedulable"])
    def test_cordon_failed(self, node):
        """If cordoning doesn't work, we raise an exception."""
        # If the node is still schedulable after the cordoning, raise a check error
        self._coreapi.patch_node.return_value = self.node_from_test_case("schedulable")
        with pytest.raises(k8s.KubernetesCheckError):
            node.cordon()
        # If the api returns an error, raise an api error
        self._coreapi.patch_node.side_effect = kubernetes.client.exceptions.ApiException("test")
        with pytest.raises(k8s.KubernetesApiError):
            node.cordon()

    @pytest.mark.parametrize("label", ["schedulable"])
    def test_cordon_dry_run(self, dry_run):
        """Cordon in dry run doesn't actually modify the node."""
        dry_run.cordon()
        self._coreapi.patch_node.assert_not_called()

    @pytest.mark.parametrize("label,has_calls", [("schedulable", False), ("unschedulable", True)])
    def test_uncordon(self, node, has_calls):
        """Cordoning a schedulable node works."""
        patch_expected = self.node_from_test_case("schedulable")
        patch_expected.spec.unschedulable = False
        self._coreapi.patch_node.return_value = patch_expected
        # Happy path, will not raise exception.
        node.uncordon()
        if has_calls:
            # Check the call to the api is what we expect.
            self._coreapi.patch_node.assert_called_with("node2", {"spec": {"unschedulable": False}})
        else:
            self._coreapi.patch_node.assert_not_called()
        assert node.is_schedulable() is True

    @pytest.mark.parametrize("label", ["unschedulable"])
    def test_uncordon_failed(self, node):
        """If uncordoning doesn't work, we raise an exception."""
        # If the node is still schedulable after the cordoning, raise a check error
        self._coreapi.patch_node.return_value = self.node_from_test_case("unschedulable")
        with pytest.raises(k8s.KubernetesCheckError):
            node.uncordon()
        # If the api returns an error, raise an api error
        self._coreapi.patch_node.side_effect = kubernetes.client.exceptions.ApiException("test")
        with pytest.raises(k8s.KubernetesApiError):
            node.uncordon()

    @pytest.mark.parametrize("label", ["unschedulable"])
    def test_uncordon_dry_run(self, dry_run):
        """Cordon in dry run doesn't actually modify the node."""
        dry_run.uncordon()
        self._coreapi.patch_node.assert_not_called()

    @pytest.mark.parametrize("label", ["schedulable"])
    def test_refresh(self, node):
        """Refreshing data works."""
        # Non-empty list
        self._coreapi.list_node.return_value = self.list_nodes("node1")
        node.refresh()
        assert "refreshed" in node._node.metadata.annotations  # pylint: disable=protected-access

    @pytest.mark.parametrize("label", ["schedulable"])
    def test_refresh_multiple_find(self, node):
        """Refreshing fails if list_node returns multiple results."""
        response = self.list_nodes("node1")
        response.items.append(self.node_from_test_case("unschedulable"))
        self._coreapi.list_node.return_value = response
        with pytest.raises(k8s.KubernetesError):
            node.refresh()

    @pytest.mark.parametrize("label", ["schedulable"])
    def test_refresh_not_found(self, node):
        """Refreshing raises an error if the node has disappeared."""
        self._coreapi.list_node.return_value = self.list_nodes("pinkunicorn")
        with pytest.raises(k8s.KubernetesError):
            node.refresh()
        self._coreapi.list_node.side_effect = kubernetes.client.exceptions.ApiException("test")
        with pytest.raises(k8s.KubernetesApiError):
            node.refresh()

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    @pytest.mark.parametrize("label", ["unschedulable"])
    def test_drain(self, mocked_wmflib_sleep, node):
        """A successful drain works as expected."""
        before_drain = mock.MagicMock()
        before_drain.items = [self.pod_from_test_case(label) for label in ["replicaset", "daemonset"]]
        # After draining, we expect the remaining pods to just be the unevictable ones.
        after_drain = mock.MagicMock()
        after_drain.items = [self.pod_from_test_case(label) for label in ["daemonset"]]
        self._coreapi.list_pod_for_all_namespaces.side_effect = [before_drain, after_drain]
        # no exception raised.
        with mock.patch("spicerack.k8s.time.sleep") as sl:
            node.drain()
            sl.assert_not_called()
        assert self._coreapi.create_namespaced_pod_eviction.call_count == 1
        assert self._coreapi.list_pod_for_all_namespaces.call_count == 2
        # no retries
        mocked_wmflib_sleep.assert_not_called()

    @pytest.mark.parametrize("label", ["unschedulable"])
    def test_drain_eventually_successful(self, node):
        """Test that retrying can eventually succeed."""
        before_drain = mock.MagicMock()
        before_drain.items = [self.pod_from_test_case(label) for label in ["replicaset", "daemonset"]]
        # After draining, we expect the remaining pods to just be the unevictable ones.
        after_drain = mock.MagicMock()
        after_drain.items = [self.pod_from_test_case(label) for label in ["daemonset"]]
        self._coreapi.list_pod_for_all_namespaces.side_effect = [before_drain, before_drain, before_drain, after_drain]
        with mock.patch("spicerack.k8s.time.sleep") as sl:
            node.drain()
            # expect sleep to be called once for max_grace_period (30s) and once from @retry (3s)
            assert sl.call_args_list == [((30,),), ((3,),)]

    @pytest.mark.parametrize("label", ["unschedulable"])
    def test_drain_leftover(self, node):
        """If there are leftover pods an exception is raised."""
        # Now assume we weren't able to evict anything
        before_drain = mock.MagicMock()
        before_drain.items = [self.pod_from_test_case(label) for label in ["replicaset", "daemonset"]]
        self._coreapi.list_pod_for_all_namespaces.return_value = before_drain
        with pytest.raises(k8s.KubernetesCheckError, match="still has 2 pods"):
            with mock.patch("spicerack.k8s.time.sleep", return_value=None) as sl:
                node.drain()
                sl.assert_called_once_with(1)

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    @pytest.mark.parametrize("label", ["unschedulable"])
    def test_drain_failed(self, mocked_wmflib_sleep, node):
        """Test draining when a pod doesn't evict."""
        # Patch get_pods to return a pod that will fail to evict
        pod_with_no_eviction = k8s.KubernetesPod(
            "bar", "foo", self._api, dry_run=False, init_obj=self.pod_from_test_case("replicaset")
        )
        pod_with_no_eviction.evict = mock.MagicMock(side_effect=k8s.KubernetesApiError("fail!"))
        node.get_pods = mock.MagicMock(return_value=[pod_with_no_eviction])

        with pytest.raises(k8s.KubernetesCheckError, match="Could not evict all pods"):
            node.drain()
            assert mocked_wmflib_sleep.call_count == 5

    @pytest.mark.parametrize("label", ["unschedulable"])
    def test_drain_dry_run(self, dry_run):
        """Draining with dry_run does nothing."""
        before_drain = mock.MagicMock()
        before_drain.items = [self.pod_from_test_case(label) for label in ["finished", "replicaset", "daemonset"]]
        # we're not actually draining, so the return value will always be the same.
        self._coreapi.list_pod_for_all_namespaces.return_value = before_drain
        with mock.patch("spicerack.k8s.time.sleep", return_value=None):
            dry_run.drain()
        self._coreapi.create_namespaced_pod_eviction.assert_not_called()

    @pytest.mark.parametrize("label", ["unschedulable"])
    def test_drain_api_error(self, node):
        """Draining with an api error will raise an exception."""
        self._coreapi.list_pod_for_all_namespaces.side_effect = kubernetes.client.exceptions.ApiException("test")
        with pytest.raises(k8s.KubernetesApiError):
            node.drain()
        self._coreapi.list_pod_for_all_namespaces.assert_called_with(field_selector="spec.nodeName=node2")

    @pytest.mark.parametrize(
        "label,expected",
        [("schedulable", []), ("tainted", [V1Taint(effect="NoExecute", key="dedicated", value="kask")])],
    )
    def test_taints(self, node, expected):
        """Test fetching taints."""
        assert node.taints == expected


class TestKubernetesPod(KubeTestBase):
    """Unit testing KubernetesPod."""

    @pytest.fixture
    def pod(self, label):
        """Obtain a test pod."""
        _pod = self.pod_from_test_case(label)
        if _pod is None:
            name = "foo"
            ns = "bar"
        else:
            name = _pod.metadata.name
            ns = _pod.metadata.namespace

        return k8s.KubernetesPod(ns, name, self._api, dry_run=False, init_obj=_pod)

    @pytest.mark.parametrize("label,expected", [("orphaned", None), ("invalid_ref", None), ("daemonset", "DaemonSet")])
    def test_controller(self, pod, expected):
        """The output of pod.controller."""
        if expected is None:
            assert pod.controller is None
        else:
            assert pod.controller.kind == expected

    @pytest.mark.parametrize("label, expected", [("daemonset", True), ("replicaset", False), ("invalid_ref", False)])
    def test_is_daemonset(self, pod, expected):
        """Detecting a daemonset."""
        assert pod.is_daemonset() is expected

    @pytest.mark.parametrize("label, expected", [("finished", True), ("replicaset", False), ("invalid_ref", False)])
    def test_is_terminated(self, pod, expected):
        """Detecting a terminated pod."""
        assert pod.is_terminated() is expected

    @pytest.mark.parametrize("label, expected", [("mirror", True), ("orphaned", False)])
    def test_is_mirror(self, pod, expected):
        """Detecting mirror pods."""
        assert pod.is_mirror() is expected

    @pytest.mark.parametrize(
        "label,expected",
        [("daemonset", False), ("replicaset", True), ("orphaned", False), ("finished", True), ("mirror", False)],
    )
    def test_is_evictable(self, pod, expected):
        """Is the pod evictable."""
        assert pod.is_evictable() is expected

    @pytest.mark.parametrize("label", ["no data"])
    def test_refresh(self, pod):
        """The requests to the api for the pod."""
        self._coreapi.read_namespaced_pod.assert_called_with("foo", "bar")
        self._api.core.assert_called_with(user="bar")
        refresher = self.pod_from_test_case("replicaset")
        self._coreapi.read_namespaced_pod.return_value = refresher
        pod.refresh()
        # Without the refresh this would fail.
        assert pod.is_evictable()
        self._coreapi.read_namespaced_pod.side_effect = kubernetes.client.exceptions.ApiException("test")
        with pytest.raises(k8s.KubernetesApiError):
            pod.refresh()

    def test_bad_init(self):
        """Test discrepancy between init object and parameters."""
        init_obj = self.pod_from_test_case("daemonset")
        with pytest.raises(k8s.KubernetesError):
            k8s.KubernetesPod("bar", "name", self._api, init_obj=init_obj)
        with pytest.raises(k8s.KubernetesError):
            k8s.KubernetesPod("namespace", "calico", self._api, init_obj=init_obj)

    def test_evict(self):
        """Evicting a pod calls the api."""
        k8s.KubernetesPod(
            "bar", "foo", self._api, init_obj=self.pod_from_test_case("replicaset"), dry_run=False
        ).evict()
        self._coreapi.create_namespaced_pod_eviction.assert_called()

    def test_evict_unevictable(self):
        """Evicting an unevictable pod raises an exception."""
        with pytest.raises(k8s.KubernetesError):
            k8s.KubernetesPod(
                "bar", "calico", self._api, init_obj=self.pod_from_test_case("daemonset"), dry_run=False
            ).evict()

    def test_evict_error(self):
        """When a pod eviction returns an error, an api error is raised."""
        self._coreapi.create_namespaced_pod_eviction.side_effect = kubernetes.client.exceptions.ApiException("test")
        with pytest.raises(k8s.KubernetesApiError):
            k8s.KubernetesPod(
                "bar", "foo", self._api, init_obj=self.pod_from_test_case("replicaset"), dry_run=False
            ).evict()
        # Also test dry run. It won't raise an exception
        k8s.KubernetesPod("bar", "foo", self._api, init_obj=self.pod_from_test_case("replicaset")).evict()

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_evict_retry(self, mocked_wmflib_sleep):
        """When a pod eviction returns 429, retry."""
        self._coreapi.create_namespaced_pod_eviction.side_effect = kubernetes.client.exceptions.ApiException(
            status=HTTPStatus.TOO_MANY_REQUESTS, reason="test"
        )
        with pytest.raises(k8s.KubernetesApiTooManyRequests):
            k8s.KubernetesPod(
                "bar", "foo", self._api, init_obj=self.pod_from_test_case("replicaset"), dry_run=False
            ).evict()
        # Check if eviction has been retried
        assert self._coreapi.create_namespaced_pod_eviction.call_count == 5
        assert mocked_wmflib_sleep.call_count == 4
        # Also test dry run. It won't raise an exception
        k8s.KubernetesPod("bar", "foo", self._api, init_obj=self.pod_from_test_case("replicaset")).evict()



"""Class API runtime_description raise cookbook."""
from spicerack.cookbook import CookbookBase, CookbookRunnerBase


class RuntimeDescriptionRaise(CookbookBase):
    """Class API runtime_description raise cookbook."""

    def get_runner(self, args):
        """As required by the parent class."""
        return ExampleRunner()


class ExampleRunner(CookbookRunnerBase):
    """As required by spicerack._cookbook."""

    @property
    def runtime_description(self):
        """As defined by the parent class."""
        raise RuntimeError("Runtime description raise")

    def run(self):
        """As required by the parent class."""



"""APT module."""
import logging
from collections.abc import Iterator
from typing import Any

from ClusterShell.MsgTree import MsgTreeElem
from ClusterShell.NodeSet import NodeSet

from spicerack.exceptions import SpicerackError
from spicerack.remote import RemoteHostsAdapter

logger = logging.getLogger(__name__)
APT_GET_ENVS: tuple[str, ...] = ("DEBIAN_FRONTEND=noninteractive",)
"""The environment variables used for all ``apt-get`` commands."""
APT_GET_INSTALL_OPTIONS: tuple[str, ...] = (
    "--quiet",
    "--yes",
    '--option Dpkg::Options::="--force-confdef"',
    '--option Dpkg::Options::="--force-confold"',
)
"""The CLI arguments passed to all ``apt-get`` commands."""
APT_GET_BASE_COMMAND: str = " ".join((*APT_GET_ENVS, "/usr/bin/apt-get", *APT_GET_INSTALL_OPTIONS))
"""The base ``apt-get`` command to execute."""


class AptGetError(SpicerackError):
    """Custom base exception class for errors in the AptGetHosts class."""


class AptGetHosts(RemoteHostsAdapter):
    """Class to manage packages via apt-get.

    Examples:
        ::

                >>> hosts = spicerack.remote().query('A:myalias')
                >>> apt_get = spicerack.apt_get(hosts)

    """

    def update(self, **kwargs: Any) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Update the list of available packages known to apt-get.

        Warnings:
            This can fail if unable to get the apt lock on the host because other apt operations are ongoing, including
            but not limited to, the periodic Puppet run that performs ``apt-get update`` before every run.
            Consider wrapping it in a :py:func:`wmflib.interactive.confirm_on_failure` call.

        Examples:
            ::

                >>> hosts = spicerack.remote().query('A:myalias')
                >>> apt_get = spicerack.apt_get(hosts)
                >>> apt_get.update()
                >>> # Optionally pass any argument accepted by run_sync()
                >>> apt_get.update(batch_size=2, print_progress_bars=False)

        Arguments:
            **kwargs: optional keyword arguments to be passed to the :py:meth:`spicerack.remote.RemoteHosts.run_sync`
                method.

        Returns:
            The result of the update operations, see :py:meth:`spicerack.remote.RemoteHosts.run_sync`.

        """
        logger.info("Running apt-get update")
        return self.run("update", **kwargs)

    def install(self, *packages: str, **kwargs: Any) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Apt-get install the provided packages.

        Install the provided packages keeping the existing configuration files (typically managed by Puppet) in a
        non-interactive way that confirms the installation of new additional binary packages (which e.g. can happen if
        a package pulls in a new dependency).

        Warnings:
            This can fail if unable to get the apt lock on the host because other apt operations are ongoing, including
            but not limited to, the periodic Puppet run that performs ``apt-get update`` before every run.
            Consider wrapping it in a :py:func:`wmflib.interactive.confirm_on_failure` call.

        Notes:
            Downgrades of package versions are not supported as they need the ``--force-yes`` CLI argument to be passed
            to ``apt-get`` and that's deemed unsafe for the possibility of unwanted results.

        Examples:
            ::

                >>> hosts = spicerack.remote().query('A:myalias')
                >>> apt_get = spicerack.apt_get(hosts)
                >>> apt_get.install('package1', 'package2')
                >>> # Optionally pass any argument accepted by run_sync()
                >>> apt_get.install('package1', 'package2', batch_size=2, print_progress_bars=False)

        Arguments:
            *packages: packages to install as positional arguments.
            **kwargs: optional keyword arguments to be passed to the :py:meth:`spicerack.remote.RemoteHosts.run_sync`
                method.

        Returns:
            The result of the installation operations, see :py:meth:`spicerack.remote.RemoteHosts.run_sync`.

        """
        if not packages:
            raise AptGetError("No packages to install were provided.")

        logger.info("Running apt-get install for the following packages: %s", packages)
        return self.run(" ".join(["install", *packages]), **kwargs)

    def run(self, apt_get_command: str, **kwargs: Any) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Execute the given apt-get command on the current hosts.

        Warnings:
            This can fail if unable to get the apt lock on the host because other apt operations are ongoing, including
            but not limited to, the periodic Puppet run that performs ``apt-get update`` before every run.
            Consider wrapping it in a :py:func:`wmflib.interactive.confirm_on_failure` call.

        Examples:
            ::

                >>> hosts = spicerack.remote().query('A:myalias')
                >>> apt_get = spicerack.apt_get(hosts)
                >>> apt_get.run('autoclean')
                >>> # Optionally pass any argument accepted by run_sync()
                >>> apt_get.run('purge package1', batch_size=2, print_progress_bars=False)

        Arguments:
            apt_get_command: the command part after apt-get to be executed (e.g. ``update``).
            **kwargs: optional keyword arguments to be passed to the :py:meth:`spicerack.remote.RemoteHosts.run_sync`
                method.

        Returns:
            The result of the update operations, see :py:meth:`spicerack.remote.RemoteHosts.run_sync`.

        """
        command = f"{APT_GET_BASE_COMMAND} {apt_get_command}"
        logger.info("Running apt-get command: %s", command)
        return self._remote_hosts.run_sync(command, **kwargs)



"""Cookbook that uses extended accessors from external modules."""
import sys

from spicerack.cookbook import CookbookBase, CookbookRunnerBase


class ExternalModulesCookbook(CookbookBase):
    """Call a Spicerack extender accessor."""

    def get_runner(self, _):
        """As required by the parent class."""
        return ExternalModulesRunner(self.spicerack)


class ExternalModulesRunner(CookbookRunnerBase):
    """The cookbook runner."""

    def __init__(self, spicerack):
        """Initialize the instance."""
        self.spicerack = spicerack

    def run(self):
        """As required by the parent class."""
        print(self.spicerack.cool_feature("Extender"), file=sys.stderr)
